
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * @see    JavaTalkServer
 * @see    JavaTalkClient
 * @author Jean-Guy Speton
 * @author Maharajapuram Balakrishnan
 *
 * This class is the thread that handles all communication with a client.
 */
class JavaTalkConnection extends Thread
{
  private static int connection_number = 0;

  protected Socket client;
  protected JavaTalkServer server;
  protected DataInputStream in;
  protected DataOutputStream out;
  protected JavaTalkUser user; // user information associated with connection

  // Initialize the streams and start the thread
  public JavaTalkConnection(Socket client_socket,
			    ThreadGroup threadgroup,
			    int priority, 
			    JavaTalkServer server) 
  {
    // Give the thread a group, a name, and a priority.
    super(threadgroup, "JavaTalkConnection-" + connection_number++);
    setPriority(priority);

    // Save our other arguments away
    client = client_socket;
    this.server = server;

    // Create the streams
    try { 
      in = new DataInputStream(client.getInputStream());
      out = new DataOutputStream(client.getOutputStream());
    }
    catch (IOException e) {
      try {
	client.close();
      }
      catch (IOException e2) {}
      System.err.println("Exception while getting socket streams: " + e);
      return;
    }

    // Obtain and store connection information.
    user = new JavaTalkUser(client.getInetAddress().getHostName(),
			    client.getPort());
    try {
      user.name = in.readLine();
    }
    catch (IOException e) {
      try {
	client.close();
      }
      catch (IOException e2) {}
      System.err.println("Exception reading user name: " + e);
      return;  // never start
    }

    server.log("CONNECT received by " + this.toString());
  }
    
  // Read messages and react to them.
  public void run()
  {
    String line;
    StringBuffer revline;
    int len;

    // The first thing we have to do is notify all other connections
    // that the user list has changed.
    broadcastAddUser();

    msgLoop: for(;;) {
      // read and parse messages.
      try {
	switch (in.readInt()) {
	  String target, sender, msg;
	  
	  // A public text message.
	case JavaTalkMessage.TEXT:
	  try {
	    sender = new String(in.readLine());
	    msg = new String(in.readLine());
	    broadcastText(sender, msg);
	  }
	  catch (IOException e) {
	    System.err.println("Text message: " + e);
	  }
	  break;

	// A private text message.
	case JavaTalkMessage.PRIV_TEXT:
	  JavaTalkConnection c;

	  try {
	    target = new String(in.readLine());
	    sender = new String(in.readLine());
	    msg = new String(in.readLine());

	    for (int i = 0; i < server.connections.size(); i++) {
	      c = (JavaTalkConnection)server.connections.elementAt(i);

	      if (target.equals(c.user.name)) {
		synchronized (c.out) {
		  c.out.writeInt(JavaTalkMessage.TEXT);
		  c.out.writeBytes("*" + sender + "*" + "\n");
		  c.out.writeBytes(msg + "\n");
		}
		System.err.println(this.getName() + " sending PRIV_TEXT " + 
				   this.user.name + " to client " +
				   c.user.name +
				   ":" + c.user.host + ":" + c.user.port);
	      }
	    }
	  }
	  catch (IOException e) {
	    System.err.println("Private text message: " + e);
	  }
	  break;

	// A graphics message.
	case JavaTalkMessage.GFX:
	  int mode, color, x1, y1, x2, y2;
	  try {
	    mode  = in.readInt();
	    color = in.readInt();
	    x1    = in.readInt();
	    y1    = in.readInt();
	    x2    = in.readInt();
	    y2    = in.readInt();
	    broadcastGfx(mode, color, x1, y1, x2, y2);
	  }
	  catch (IOException e) {
	    System.err.println("Gfx message: " + e);
	  }
	  break;

	// A disconnect message.
	case JavaTalkMessage.DISCONNECT:
	  server.log("DISCONNECT received by " + this);
	  broadcastDelUser();
	  break msgLoop;

	} // switch
      }
      catch (EOFException e) {
	server.log(this + " -- EOF received, closing socket.");
	break msgLoop;
      }
      catch (IOException e) {
	System.err.println("Exception in msgLoop: " + e);
      }
    }

    try {
      client.close();
    }
    catch (IOException e2) {}

    // Remove ourselves from the connections list.
    synchronized (server.connections) {
      for(int i = 0; i < server.connections.size(); i++) {
	JavaTalkConnection c;
	c = (JavaTalkConnection)server.connections.elementAt(i);
	// if the connection thread isn't alive anymore, 
	// remove it from the Vector and List.
	if (c == this) {
	  server.log("Removing connection " +
		     server.connections.elementAt(i));
	  server.connections.removeElementAt(i);
	  break;
	}
      }
    }
    

  }

  // First of all send an ADD_USER for pre-existing clients to this' client.
  // Then sends an ADD_USER message to every client, including this,
  // with this client's name, host and port.
  private void broadcastAddUser()
  {
    JavaTalkConnection c;

    // send our client existing user list, minus itself
    synchronized (out) {
      for (int i = 0; i < server.connections.size(); i++) {
	c = (JavaTalkConnection)server.connections.elementAt(i);
	if (c != this && c.isAlive()) {
	  try {
	    out.writeInt(JavaTalkMessage.NEW_USER);
	    out.writeBytes(c.user.name + "\n" +
				c.user.host + "\n");
	    out.writeInt(c.user.port);
	    server.log(getName() + " sending NEW_USER " + c.user.name +
		       " to client " + user.name + ":" +
		       user.host + ":" + user.port);
	  }
	  catch (IOException e) {}
	}
      }
    }

    // synchronize on the connections list
    synchronized (server.connections) {
      for (int i = 0; i < server.connections.size(); i++) {
	c = (JavaTalkConnection)server.connections.elementAt(i);

	// ensure no one else writes to socket at same time
	synchronized (c.out) {
	  try {
	    c.out.writeInt(JavaTalkMessage.ADD_USER);
	    c.out.writeBytes(this.user.name + "\n" +
			     this.user.host + "\n");
	    c.out.writeInt(this.user.port);
	    server.log(getName() + " sending ADD_USER " + user.name +
		       " to client " + c.user.name + ":" +
		       c.user.host + ":" + c.user.port);
	  }
	  catch (IOException e) {}
	}
      }
    }
  } 

  // Sends a DEL_USER message to every client, excluding this,
  // with this client's name, host and port.
  private void broadcastDelUser()
  {
    // synchronize on the connections list
    synchronized (server.connections) {
      JavaTalkConnection c;
     
      for (int i = 0; i < server.connections.size(); i++) {
	c = (JavaTalkConnection)server.connections.elementAt(i);

	// No point in notifying our client.
	if (c != this) {
	  // ensure no one else writes to socket at same time
	  synchronized (c.out) {
	    try {
	      c.out.writeInt(JavaTalkMessage.DEL_USER);
	      c.out.writeBytes(this.user.name + "\n");
	      server.log(getName() + " sending DEL_USER " + 
			 user.name + " to client " + c.user.name +
			 ":" + c.user.host + ":" + c.user.port);
	    }
	    catch (IOException e) {}
	  }
	}
      }
    }
  } 

  // Sends a TEXT message to every client, including this,
  // with the sending client's name and message.
  private void broadcastText(String sender, String msg)
  {
    // synchronize on the connections list
    synchronized (server.connections) {
      JavaTalkConnection c;
     
      for (int i = 0; i < server.connections.size(); i++) {
	c = (JavaTalkConnection)server.connections.elementAt(i);

	// ensure no one else writes to socket at same time
	synchronized (c.out) {
	  try {
	    c.out.writeInt(JavaTalkMessage.TEXT);
	    c.out.writeBytes(sender + "\n" + msg + "\n");
	    server.log(getName() + " sending TEXT " + 
		       sender + " to client " + c.user.name +
		       ":" + c.user.host + ":" + c.user.port);
	  }
          catch (IOException e) {
	    System.err.println("broadcastText: " + e);
	  }
	}
      }
    }
  }

  // Sends a GFX message to every client, including this,
  // with the sending client's name and message.
  private void broadcastGfx(int mode, int color, int x1, int y1,
			      int x2, int y2)
  {
    // synchronize on the connections list
    synchronized (server.connections) {
      JavaTalkConnection c;
     
      for (int i = 0; i < server.connections.size(); i++) {

	if ((c = (JavaTalkConnection)server.connections.elementAt(i)) != this) {

	  // ensure no one else writes to socket at same time
	  synchronized (c.out) {
	    try {
	      c.out.writeInt(JavaTalkMessage.GFX);
	      c.out.writeInt(mode);
	      c.out.writeInt(color);
	      c.out.writeInt(x1);
	      c.out.writeInt(y1);
	      c.out.writeInt(x2);
	      c.out.writeInt(y2);
	      server.log(getName() + " sending GFX " + 
			 mode + " to client " + c.user.name +
			 ":" + c.user.host + ":" + c.user.port);
	    }
	    catch (IOException e) {
	      System.err.println("broadcastGfx: " + e);
	    }
	  }
	}
      }
    }
  }

  // This method returns the string representation of the Connection.
  public String toString() {
    return "[" + getName() + " connected to: " +
      client.getInetAddress().getHostName() + ":" + client.getPort() + "]";
  }
}
